% Gonzalo Vaca-Castano.
% University of Central Florida
% 2012

% Obtain the SIFT interest points, and calculate the SIFT descriptor.
% INPUTS
%   image: Image to look for interest points. The image is in gray scale.
%   num_octaves: Number of octaves
%   num_intervals: Number of intervals s in to achieve the double of the sigma
%   value. In the original paper s=2
%   sigmavalue: Value of the initial sigma
% OUTPUTS
%   sift_desc: matrix containing in each row an intersting point
%               [x1 y1  octave  num_imagen_in_the_scale  sigma_of_the_scale  rientation(bins) ]  
function [sift_desc]=sift(image,num_octaves,num_intervals,sigmavalue)


[dimx,dimy]=size(image);

%define constants of the application
antialiassigma=0.5;
k1=2^(1/num_intervals);  
r_curvature=5;
contrast_threshold=0.03; %Threshold determining a low contrast keypoint to be eliminated

if sigmavalue>1.1
 % contrast_threshold=0.07;
end
if sigmavalue==1.4
    contrast_threshold=0.06;
end

updown=1;


%normalize image input
% Assuming 8 bits image
image1=double(image)/255;


% %%%%%%% Allocate memory %%%%%%%%%
gaussians = cell(num_intervals+3,num_octaves);
dogs=cell(num_intervals+2,num_octaves);
magnitude=cell(num_intervals,num_octaves);
orientation=cell(num_intervals,num_octaves);
for i=1:num_octaves
   for j=1:num_intervals+3
      gaussians(j,i)= {zeros(dimx/(2^(i-2)),dimy/(2^(i-2)))};
   end 
   for j=1:num_intervals+2
      dogs(j,i)= {zeros(dimx/(2^(i-2)),dimy/(2^(i-2)))};
   end
   for j=1:num_intervals
       magnitude(j,i)=  {zeros((dimx/(2^(i-2)))-2,(dimy/(2^(i-2)))-2)};
       orientation(j,i)={zeros((dimx/(2^(i-2)))-2,(dimy/(2^(i-2)))-2)};
   end
end



% %%% Create first interval of the first octave %%%%%
init_image=impyramid(gauss_filter(image1,antialiassigma,4*antialiassigma),'expand');
gaussians(1)={gauss_filter(init_image,sigmavalue,4*sigmavalue)};




% %%% Generates all the blurred out images for each octave %%%%
% %%%               and the DoG images                     %%%%
for i=1:num_octaves
   sigma=sigmavalue;    %reset the sigma value   
   for j=1:(num_intervals+2)
       sigma=sigma*2^((j-1)/2);     %Assign a sigma value acording to the scale 
       previmage=cell2mat(gaussians(j,i));  %Obtain the previous image
       newimage=gauss_filter(previmage,sigma,4*sigma);  %apply a new smoothing       
       dog=previmage-newimage;  %calculate the difference of gaussians
       
       %save the results
       gaussians(j+1,i)={newimage};
       dogs(j,i)={dog};
   end
   
   %Build the  init image in the next level
   if(i<num_octaves)
       lowscale=cell2mat(gaussians(num_intervals+1,i));
       upscale=impyramid(lowscale,'reduce');
       gaussians(1,i+1)={upscale};
   end    
end



% %%%%% Find the local extrema (Maximum o minimum) %%%%%%%%%%
% %%%%%%%%%%% List Interest points %%%%%%%%%%%%%%%%%%%%%%%%%%
list_points=[];
for i=1:num_octaves
    for j=2:(num_intervals+1)
        % Obtain the matrices where to look for the extrema
        level=cell2mat(dogs(j,i));
        up=cell2mat(dogs(j+1,i));
        down=cell2mat(dogs(j-1,i));
        
        [sx,sy]=size(level);
        
        %look for a local maxima
        local_maxima=(level(2:sx-1,2:sy-1)>level(1:sx-2,1:sy-2)) & ( level(2:sx-1,2:sy-1) > level(1:sx-2,2:sy-1) ) & (level(2:sx-1,2:sy-1)>level(1:sx-2,3:sy)) & (level(2:sx-1,2:sy-1)>level(2:sx-1,1:sy-2)) & (level(2:sx-1,2:sy-1)>level(2:sx-1,3:sy)) & (level(2:sx-1,2:sy-1)>level(3:sx,1:sy-2)) & (level(2:sx-1,2:sy-1)>level(3:sx,2:sy-1)) & (level(2:sx-1,2:sy-1)>level(3:sx,3:sy)) ;
        local_maxima=local_maxima & (level(2:sx-1,2:sy-1)>up(1:sx-2,1:sy-2)) & ( level(2:sx-1,2:sy-1) > up(1:sx-2,2:sy-1) ) & (level(2:sx-1,2:sy-1)>up(1:sx-2,3:sy)) & (level(2:sx-1,2:sy-1)>up(2:sx-1,1:sy-2)) & (level(2:sx-1,2:sy-1)>up(2:sx-1,2:sy-1)) & (level(2:sx-1,2:sy-1)>up(2:sx-1,3:sy)) & (level(2:sx-1,2:sy-1)>up(3:sx,1:sy-2)) & (level(2:sx-1,2:sy-1)>up(3:sx,2:sy-1)) & (level(2:sx-1,2:sy-1)>up(3:sx,3:sy)) ;
        local_maxima=local_maxima & (level(2:sx-1,2:sy-1)>down(1:sx-2,1:sy-2)) & ( level(2:sx-1,2:sy-1) > down(1:sx-2,2:sy-1) ) & (level(2:sx-1,2:sy-1)>down(1:sx-2,3:sy)) & (level(2:sx-1,2:sy-1)>down(2:sx-1,1:sy-2)) & (level(2:sx-1,2:sy-1)>down(2:sx-1,2:sy-1)) & (level(2:sx-1,2:sy-1)>down(2:sx-1,3:sy)) & (level(2:sx-1,2:sy-1)>down(3:sx,1:sy-2)) & (level(2:sx-1,2:sy-1)>down(3:sx,2:sy-1)) & (level(2:sx-1,2:sy-1)>down(3:sx,3:sy)) ;
        
        %look for a local minima
        local_minima=(level(2:sx-1,2:sy-1)>level(1:sx-2,1:sy-2)) & ( level(2:sx-1,2:sy-1) > level(1:sx-2,2:sy-1) ) & (level(2:sx-1,2:sy-1)>level(1:sx-2,3:sy)) & (level(2:sx-1,2:sy-1)>level(2:sx-1,1:sy-2)) & (level(2:sx-1,2:sy-1)>level(2:sx-1,3:sy)) & (level(2:sx-1,2:sy-1)>level(3:sx,1:sy-2)) & (level(2:sx-1,2:sy-1)>level(3:sx,2:sy-1)) & (level(2:sx-1,2:sy-1)>level(3:sx,3:sy)) ;
        local_minima=local_minima & (level(2:sx-1,2:sy-1)>up(1:sx-2,1:sy-2)) & ( level(2:sx-1,2:sy-1) > up(1:sx-2,2:sy-1) ) & (level(2:sx-1,2:sy-1)>up(1:sx-2,3:sy)) & (level(2:sx-1,2:sy-1)>up(2:sx-1,1:sy-2)) & (level(2:sx-1,2:sy-1)>up(2:sx-1,2:sy-1)) & (level(2:sx-1,2:sy-1)>up(2:sx-1,3:sy)) & (level(2:sx-1,2:sy-1)>up(3:sx,1:sy-2)) & (level(2:sx-1,2:sy-1)>up(3:sx,2:sy-1)) & (level(2:sx-1,2:sy-1)>up(3:sx,3:sy)) ;
        local_minima=local_minima & (level(2:sx-1,2:sy-1)>down(1:sx-2,1:sy-2)) & ( level(2:sx-1,2:sy-1) > down(1:sx-2,2:sy-1) ) & (level(2:sx-1,2:sy-1)>down(1:sx-2,3:sy)) & (level(2:sx-1,2:sy-1)>down(2:sx-1,1:sy-2)) & (level(2:sx-1,2:sy-1)>down(2:sx-1,2:sy-1)) & (level(2:sx-1,2:sy-1)>down(2:sx-1,3:sy)) & (level(2:sx-1,2:sy-1)>down(3:sx,1:sy-2)) & (level(2:sx-1,2:sy-1)>down(3:sx,2:sy-1)) & (level(2:sx-1,2:sy-1)>down(3:sx,3:sy)) ;
           
        extrema=local_maxima | local_minima;
        
        %indices of the extrema points
        [x,y]=find(extrema);
        numtimes=size(find(extrema));
        
        for k=1:numtimes
            x1=x(k);
            y1=y(k);
            if(abs(level(x1+1,y1+1))<contrast_threshold)    %low contrast point are discarded
               extrema(x1,y1)=0;
            else   %keep being extrema, check for edge
               rx=x1+1;
               ry=y1+1;
               fxx= level(rx-1,ry)+level(rx+1,ry)-2*level(rx,ry);   % double derivate in x direction
               fyy= level(rx,ry-1)+level(rx,ry+1)-2*level(rx,ry);   % double derivate in y direction
               fxy= level(rx-1,ry-1)+level(rx+1,ry+1)-level(rx-1,ry+1)-level(rx+1,ry-1); %derivate inx and y direction
               trace=fxx+fyy;
               deter=fxx*fyy-fxy*fxy;
               curvature=trace*trace/deter;
               curv_threshold= ((r_curvature+1)^2)/r_curvature;
               if(deter<0 || curvature>curv_threshold)   %Reject edge points
                   extrema(x1,y1)=0;
               end
            end
        end
        
        [x,y]=find(extrema);
        numtimes=size(find(extrema));
        
        %check
        list_points=[list_points; [x y repmat(i,numtimes,1) repmat(j-1,numtimes,1) zeros(numtimes,1) zeros(numtimes,1)] ];
        
    end
end


% %%%%  generate an orientation for all the key points  %%%%%%%%%%%%
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% calculate magnitude,orientation
for i=1:num_octaves
    for j=1:num_intervals
        previmage=cell2mat(gaussians(j+updown,i));   %add 2 to associate to the sigma of the upper image, or add 1 to associate to the sigma of the lower one.
        [rx,ry]=size(previmage);
        dx=previmage(3:rx,2:ry-1)-previmage(1:rx-2,2:ry-1);
        dy=previmage(2:rx-1,1:ry-2)-previmage(2:rx-1,3:ry);
        mag=sqrt(dx.*dx + dy.*dy);
        ori=atan2(dy,dx);
        magnitude(j,i)={mag};
        orientation(j,i)={ori};        
    end
end

new_ones=[];

% Calculate the orientation of the keypoints
points_size=size(list_points,1);
for i=1:points_size     %Look in the keypoints
    octave=list_points(i,3);    %octave of the interest point
    level=(list_points(i,4))+updown;    %level of the interest point
    x1=list_points(i,1);    % x position of the interest point
    y1=list_points(i,2);    % y position of the interest point
    sigma=sigmavalue*k1^(level-1);  %calculate the sigma in the corresponding scale
    list_points(i,5)=sigma;
    mag=cell2mat(magnitude(level-updown,octave));   %magnitude of the gradient
    ori=cell2mat(orientation(level-updown,octave)); %direction of the gradient
    [weights,n]=gausswindow(1.5*sigma); %create weigths around the interest point
    hist=zeros(36,1);   %create an empty histogram
    
    [xmax,ymax]=size(mag);
    a11=x1-n;
    a12=x1+n;
    
    a21=y1-n;
    a22=y1+n;
   
    a31=1;
    a32=2*n+1;
    
    a41=1;
    a42=2*n+1;
    if((x1-n)<1)
        a11=1;
       % a31=(n-x1);
       % if(n==x1)
           a31=n-x1+2; 
       % end
    end
    
    if((y1-n)<1)
        a21=1;
       % a41=n-y1;
       % if(n==y1)
            a41=n-y1+2;
       % end
    end
    
    if((x1+n)>xmax)  
        a12=xmax;
        a32=(2*n+1)-(x1+n-xmax);
    end
    
    if((y1+n)>ymax)  
        a22=ymax;
        a42=(2*n+1)-(y1+n-ymax);
    end
        
    %w=mag(a1,a2).*weights(a3,a4);    %weight of each pixel in the bin, after gaussian weighting and magnitude assignation.    
    %ori=ori(a1,a2);
    w=mag(a11:a12,a21:a22).*weights(a31:a32,a41:a42);    %weight of each pixel in the bin, after gaussian weighting and magnitude assignation.    
    ori=ori(a11:a12,a21:a22);
    
    ori=ceil((ori+pi)*18/pi);   %obtain the direction separated in bins of each one of the points near to the keypoint
    
    [orix,oriy]=size(ori);
    %calculate histogram
    for i1=1:orix
        for j1=1:oriy
            hist(ori(i1,j1))=hist(ori(i1,j1))+w(i1,j1);
        end
    end
    
    %calculate histogram
%     for i1=1:(2*n+1)
%         for j1=1:(2*n+1)
%             hist(ori(i1,j1))=hist(ori(i1,j1))+w(i1,j1);
%         end
%     end
%     
    [maxval,bin]=max(hist);
    list_points(i,6)=bin;
    lista=find(hist>(0.8*maxval));
    more=size(lista,1);
    if(more>1)
        for i3=1:more
            if(lista(i3) ~= bin)
                new_ones=[new_ones;  [x1 y1 octave list_points(i,4) list_points(i,5) lista(i3)]];
            end
        end
    end
    
    
end

%sift_desc=[list_points ; new_ones];
sift_desc=[list_points];


%  figure; imshow(image,[]);hold on;
%      for j=1:size(list_points,1),
%          plot( list_points(j,2)*2^(list_points(j,3)-2), list_points(j,1)*2^(list_points(j,3)-2), 'ro', 'color', [1 0 0], 'markerfacecolor', [1 0 0], 'markersize', 2.5 );
%      end
%      hold off
%      
    
  list_points= sift_desc;
  figure; imshow(image,[]);hold on;
  x3=[];
  y3=[];
      for j=1:size(list_points,1),
          x4=list_points(j,2)*2^(list_points(j,3)-2);
          y4=list_points(j,1)*2^(list_points(j,3)-2);
          plot(x4 ,y4 , 'ro', 'color', [1 0 0], 'markerfacecolor', [1 0 0], 'markersize', 2.5 );
          x3=[x3;x4];
          y3=[y3;y4];
      end
      hold off

%list_points= sift_desc;
dire=((list_points(:,6)/18))*pi + pi/2;
%dire=((list_points(:,6)-18)/18)*pi ;
mag=(list_points(:,3))+list_points(:,4);

locs=[y3 x3 mag dire];
  showkeys(image, locs);
% imshow(I);
%plot(locs(:,2),locs(:,1),'r.');


end